#include <stdio.h>
#include "julia.h"
#include <math.h>
#include <unistd.h>

void Julia::render(double c_re, double c_im, double radius, double aspect, int dwell, int width, int colouring)
{
  double xmin, xmax, ymin, ymax;
  int height;
  double re = 0, im = 0;

  if (aspect > 1) {
    ymin = im - radius;
    ymax = im + radius;
    xmin = re - (radius*aspect);
    xmax = re + (radius*aspect);
    height = (int)(width/aspect);
  } else {
    xmin = re - radius;
    xmax = re + radius;
    ymin = im - (radius/aspect);
    ymax = im + (radius/aspect);
    height = (int)(width*aspect);
  }

  fprintf(stderr, "Rendering Julia fractal for:\n"
	  "\t%.2f < re < %.2f, %.2f < im < %.2f, dwell limit = %d\n"
	  "image size: %dx%d\n",
	  xmin, xmax, ymin, ymax, dwell, width, height);

  char header[128];
  sprintf(header, "P6\n# CREATOR Fractalgen, Jonas Norling\n%d %d\n255\n", width, height);
  write(1, header, strlen(header));


  //  unsigned char* data = new unsigned char[width*height*3]; //Tokstort
  int data;
  int colour[3];
  int i;
  double a, b, e;
  
  for (int yp=0; yp < height; yp++) {
    for (int xp=0; xp < width; xp++) {
      a = ((double)xp/width) * (xmax-xmin) + xmin;
      b = (1-(double)yp/height) * (ymax-ymin) + ymin;


      for (i=1; i < dwell; i++) {
	e = a;
	a = a*a - b*b;
	b = 2*e*b;
	a+=c_re;
	b+=c_im;
	if (sqrt(a*a + b*b) > 2)
	  break;
      }
      

      switch (colouring) {
      case COLOUR:
	if (i >= dwell) {
	  colour[0] = colour[1] = colour[2] = 0;
	} else {
	  //A colour scheme of black->red->orange->yellow->white is used
	  colour[0] = (int)((double)i/dwell * 255 * 3);
	  colour[1] = (int)((double)i/dwell * 255 * 3/2);
	  colour[2] = (int)((double)i/dwell * 255 * 1);
	  if (colour[0] > 255) colour[0] = 255;
	  if (colour[1] > 255) colour[1] = 255;
	  if (colour[2] > 255) colour[2] = 255;
	}
	
	data = (255 << 24) | (colour[0] << 16) | (colour[1] << 8) | (colour[2]);
	break;
      case BW:
	if (i >= dwell)
	  data = 0xFFFFFF;
	else
	  data = 0;
	break;
      }
      write(1, &data, 3);
    }
  }



}
