// Copyright Jonas Norling 2000-2002
// <jonas@jomato.se>
// GNU GPL applies.

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>


class colour
{
public:
  u_char rgb[3];

  colour()
  {
    rgb[0] = rgb[1] = rgb[2] = 0;
  }

  void set(int c)
  {
    // BYTEORDER DEPENDANT!
    rgb[0] = (c >> 16) & 0xFF;
    rgb[1] = (c >> 8) & 0xFF;
    rgb[2] = c & 0xFF;
  }

  void set(colour &c)
  {
    rgb[0] = c.rgb[0];
    rgb[1] = c.rgb[1];
    rgb[2] = c.rgb[2];
  }

  char* toString()
  {
    char* rtn = new char[8];
    sprintf(rtn, "#%.2x%.2x%.2x", rgb[0], rgb[1], rgb[2]);
    return (rtn);
  }

  void fade(float f)
  {
    rgb[0] -= (u_char)(rgb[0] * f);
    rgb[1] -= (u_char)(rgb[1] * f);
    rgb[2] -= (u_char)(rgb[2] * f);
  }

  int get()
  {
    int i = 0;
    i |= rgb[0];
    i <<= 8;
    i |= rgb[1];
    i <<= 8;
    i |= rgb[2];
    return (i);
  }
};


class colourmap
{
public:
  colour* colours;

  colourmap(int count) {
    fprintf(stderr, "Generating colour map of %d colours.\n", count);

    colours = (colour*) calloc(count + 1, sizeof(colour));
    
    if (count <= 8) {
      colours[0].set(0x0);
      colours[1].set(0xff8888);
      colours[2].set(0x77dd77);
      colours[3].set(0x8888ff);
      colours[4].set(0xff88ff);
      colours[5].set(0x88ffff);
      colours[6].set(0xffff88);
      colours[7].set(0x3cc1c1);
      colours[8].set(0xb56e19);
    } else {
      colours[0].set(0x0);
      colours[1].set(0xff8888);
      colours[2].set(0x77dd77);
      colours[3].set(0x8888ff);
      colours[4].set(0xff88ff);
      colours[5].set(0x88ffff);
      colours[6].set(0xffff88);
      colours[7].set(0x3cc1c1);
      colours[8].set(0xb56e19);
      for (int i=9; i < count; i++) {
	colours[i].set(0xffffff);
	colours[i].fade((float)i/count);
      }
    }
  }
};


class complex
{
public:
  double re, im;
  complex(double _re, double _im)
  {
    re = _re;
    im = _im;
  }

  complex()
  {
    re = im = 0;
  }

  complex(complex* a)
  {
    re = a->re;
    im = a->im;
  }

  void set(double _re, double _im)
  {
    re = _re;
    im = _im;
  }

  void set(complex a)
  {
    re = a.re;
    im = a.im;
  }

  void add(complex a)
  {
    re += a.re;
    im += a.im;
  }

  complex operator*(complex a)
  {
    complex rtn;
    rtn.set((re*a.re - im*a.im), (re*a.im + im*a.re));

    return rtn;
  }

  complex operator*(double a)
  {
    complex rtn;
    rtn.set(re*a, im*a);
    return (rtn);
  }

  complex operator+(complex a)
  {
    complex rtn;
    rtn.set(re+a.re, im+a.im);
    return (rtn);
  }

  complex operator-(complex a)
  {
    complex rtn;
    rtn.set(re-a.re, im-a.im);
    return (rtn);
  }

  complex operator+(double a)
  {
    complex rtn;
    rtn.set(re+a, im+a);
    return (rtn);
  }

  complex operator/(complex a)
  {
    complex rtn;
    rtn.set(
	    (re*a.re + im*a.im)
	             /
	    (a.re*a.re + a.im*a.im),
	    
	    (im*a.re - re*a.im)
	             /
	    (a.re*a.re + a.im*a.im));
    return (rtn);
  }

  char* toString()
  {
    char* str = new char[16];
    sprintf(str, "(%.3f + %.3fi)", re, im);
    return (str);
  }

  void parseString(char* str)
  {
    sscanf(str, "%lf,%lf", &re, &im);
  }

  double abs()
  {
    return (sqrt(re*re + im*im));
  }

  complex negate()
  {
    complex res;
    res.set(-re, -im);
    return(res);
  }
};


class function
{
public:
  complex eval(complex x);
  
  char* toString()
  {
    return (NULL);
  }
};

class staticfunc : public function
{
public:
  complex eval(complex x)
  {
    complex a;
    a.set( (x*x*x*x) + (x*x*x) + x + (-1.0) );
    return (a);
  }
};

class staticderiv : public function
{
public:
  complex eval(complex x)
  {
    complex a;
    a.set( (x*x*x)*4.0 + (x*x)*3.0 + 1.0 );
    return (a);
  }
};

class polynom : public function
{
public:
  int grad;
  complex** koeff;

  polynom()
  {
    grad = 0;
    koeff = new complex*[grad+1];
    koeff[0] = new complex(1, 0);
  }

  polynom(complex koeff_1, complex koeff_0)
  {
    grad = 1;
    koeff = new complex*[grad+1];
    koeff[0] = new complex(&koeff_0);
    koeff[1] = new complex(&koeff_1);
  }
  
  polynom(polynom* poly)
  {
    grad = poly->grad;
    koeff = new complex*[grad+1];
    for (int i=0; i<=grad; i++) {
      koeff[i] = new complex(poly->koeff[i]);
    }
  }

  void multiply(polynom* poly)
  {
    int newg = grad + poly->grad;
    complex** newk = new complex*[newg+1];
    bzero(newk, (newg+1)*sizeof(complex*));

    int i,j;

    for (i=0; i<=grad; i++) {
      for (j=0; j<=poly->grad; j++) {
	if (newk[i+j] == NULL)
	  newk[i+j] = new complex(0, 0);
	printf("newk[%d] = %d\n", i+j, newk[i+j]);
	newk[i+j]->add(     *(koeff[i]) * *(poly->koeff[j])      );
      }
    }

    delete koeff;
    koeff = newk;
    grad = newg;
  }

  void differentiate()
  {
    for (int i=0; i < grad; i++) {
      koeff[i]->set(   *(koeff[i+1]) * (i+1)    );
    }
    koeff[grad]->set(0,0);
  }

  complex eval(complex x)
  {
    complex res;
    for (int i=0; i <= grad; i++) {
      complex tmp;
      tmp.set(1, 0);
      for (int c=0; c < i; c++) {
	tmp.set(tmp * x);
      }
      tmp.set(tmp * *koeff[i]);
      res.add(tmp);
    }
    return (res);
  }

  char* toString()
  {
    char* str = new char[128];
    str[0] = '\0';
    for (int i=grad; i>=0; i--) {
      char buff[32];
      sprintf(buff, " + %sx^%d", koeff[i]->toString(), i);
      strcat(str, buff);
    }
    return (str);
  }
};


void newton(complex p, int rootc, complex** roots, function* func, function* deriv, int dwell, float rootsize, float ringwidth, colour& c, colourmap* map, bool paintiter)
{
  int iter=0;
  complex res;
  res.set(p);
  static float s;
  while (iter < dwell) {
    for (int i=0; i<rootc; i++) {
      s = (res - *roots[i]).abs();
      if (s < rootsize) {
	if (iter==0 && s>(rootsize-ringwidth)) {
	  if (!paintiter) {
	    c.set(map->colours[i+1]);
	    c.fade(0.3);
	  } else {
	    c.set(map->colours[1]);
	  }
	} else {
	  if (!paintiter) {
	    c.set(map->colours[i+1]);
	    c.fade((float)iter/dwell);
	  } else {
	    c.set(map->colours[iter]);
	  }
	}
	return;
      }
    }
    res.set(res - (((polynom*)func)->eval(res)) / (((polynom*)deriv)->eval(res)) );
	//res.set(res - (((staticfunc*)func)->eval(res)) / (((staticderiv*)deriv)->eval(res)) );
    iter++;
  }
  c.set(0x0);
}


void printhelp(char* error)
{
  fprintf(stderr, "%s\n\n", error);

  fprintf(stderr, "%s <root> [<root> [...]] <re> <im> <radius> <as> <dwell> <width> <rsize> <itercolour>\n", "newton");

  fprintf(stderr, "A PNM-image is written to stdout\n\n");

  fprintf(stderr, "roots:\tEnumeration of the equation's roots in the format re,im\n"
	  "re:\tThe center point of the fractal, real axis coordinate\n"
	  "im:\tThe center point of the fractal, imaginary axis coordinate\n"
	  "radius:\tThe size of the fractalm given as a radius from the center point\n"
	  "as:\tAspect ratio. 1.25 is advised\n"
          "dwell:\tDwell limit. Give up after this number of iterations\n"
          "width:\tImage width in pixels\n"
	  "rsize:\tMinimum distance between roots\n"
	  "itercolour:\tIf \"1\", points will be coloured according to the number of iterations needed for them to reach a root\n"
          "\n");

  fprintf(stderr, "Example:\n"
          "\tnewton -2,0 0,0 2,0 0 0 2 1.25 20 640 0.1 0\n"
          "\n");

  fprintf(stderr, "Copyright 2000-2002 Jonas Norling. GNU GPL applies\n");
}


void renderNewton(double re, double im, double radius, double aspect, int dwell, int width, int rootc, complex** roots, float rootsize, bool paintiter)
{
  double xmin, xmax, ymin, ymax;
  int height;

  if (aspect > 1) {
    ymin = im - radius;
    ymax = im + radius;
    xmin = re - (radius*aspect);
    xmax = re + (radius*aspect);
    height = (int)(width/aspect);
  } else {
    xmin = re - radius;
    xmax = re + radius;
    ymin = im - (radius/aspect);
    ymax = im + (radius/aspect);
    height = (int)(width/aspect);
  }

  float pixelwidth = fabs((xmax-xmin) / width);

  fprintf(stderr, "Rendering Newton-Raphson fractal for:\n"
	  "\t%.2f < re < %.2f, %.2f < im < %.2f, dwell limit = %d\n"
	  "\timage size: %dx%d. Every pixel is %f wide.\n\n",
	  xmin, xmax, ymin, ymax, dwell, width, height, pixelwidth);



  function* poly = new polynom();
  for (int i=0; i < rootc; i++) {
    polynom* tmp = new polynom(*(new complex(1.0, 0.0)), roots[i]->negate());
    ((polynom*)poly)->multiply(tmp);
  }

  function* deriv = new polynom((polynom*)poly);
  ((polynom*)deriv)->differentiate();

  fprintf(stderr, "Given roots yield this polynominal:\n"
	  "\t%s\n"
	  "Differentiated it is:\n"
	  "\t%s\n", ((polynom*)poly)->toString(), ((polynom*)deriv)->toString());


  //  poly = new staticfunc;
  //  deriv = new staticderiv;

  char header[128];
  sprintf(header, "P6\n# CREATOR Fractalgen, Jonas Norling\n%d %d\n255\n", width, height);
  write(1, header, strlen(header));

  
  int data;
  int i;
  double p_re, p_im;
  int to_root;
  complex p;
  int n;
  colour c;
  colourmap *map;

  if (paintiter) {
    map = new colourmap(dwell);
  } else {
    map = new colourmap(rootc);
  }

  for (int yp=0; yp < height; yp++) {
    for (int xp=0; xp < width; xp++) {
      p_re = ((double)xp/width) * (xmax-xmin) + xmin;
      p_im = (1-(double)yp/height) * (ymax-ymin) + ymin;
      p.set(p_re, p_im);

      newton(p, rootc, roots, poly, deriv, dwell, rootsize, pixelwidth*2, c, map, paintiter);
      n = c.get();
      write(1, ((char*)&n)+2, 1);
      write(1, ((char*)&n)+1, 1);
      write(1, ((char*)&n)+0, 1);
    }
  }

}


int main(int argc, char* argv[])
{
  double re, im, ra, as, rootsize;
  int dwell, width;
  int rootcount = 0;
  complex** roots;
  bool itercolour;

  if (argc < 10) {
    printhelp("Insufficient number of arguments");
    exit(2);
  }

  rootcount = argc - 9;

  fprintf(stderr, "Parsing %d roots:\n", rootcount);
  roots = new complex*[rootcount];
  for (int i=0; i < rootcount; i++) {
    roots[i] = new complex();
    roots[i]->parseString(argv[i+1]);
    fprintf(stderr, "\t%s\tas %s\n", argv[i+1], roots[i]->toString());
  }

  re = strtod(argv[rootcount+1], NULL);
  im = strtod(argv[rootcount+2], NULL);
  ra = strtod(argv[rootcount+3], NULL);
  as = strtod(argv[rootcount+4], NULL);
  dwell = atoi(argv[rootcount+5]);
  width = atoi(argv[rootcount+6]);
  rootsize = strtod(argv[rootcount+7], NULL);
  itercolour = atoi(argv[rootcount+8]);

  renderNewton(re, im, ra, as, dwell, width, rootcount, roots, rootsize, itercolour);
}


